@@ -1,27 +1,10 @@
-// Copyright, 2021, by Samuel G. D. Williams. <http://www.codeotaku.com>
-// 
-// Permission is hereby granted, free of charge, to any person obtaining a copy
-// of this software and associated documentation files (the "Software"), to deal
-// in the Software without restriction, including without limitation the rights
-// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-// copies of the Software, and to permit persons to whom the Software is
-// furnished to do so, subject to the following conditions:
-// 
-// The above copyright notice and this permission notice shall be included in
-// all copies or substantial portions of the Software.
-// 
-// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-// THE SOFTWARE.
+// Released under the MIT License.
+// Copyright, 2021-2025, by Samuel Williams.
 
 #include "kqueue.h"
 #include "selector.h"
-#include "list.h"
-#include "array.h"
+#include "../list.h"
+#include "../array.h"
 
 #include <sys/event.h>
 #include <sys/ioctl.h>
@@ -43,8 +26,6 @@ enum {
 #define IO_EVENT_SELECTOR_KQUEUE_USE_INTERRUPT
 #endif
 
-static VALUE IO_Event_Selector_KQueue = Qnil;
-
 enum {KQUEUE_MAX_EVENTS = 64};
 
 // This represents an actual fiber waiting for a specific event.
@@ -175,7 +156,7 @@ size_t IO_Event_Selector_KQueue_Type_size(const void *_selector)
 }
 
 static const rb_data_type_t IO_Event_Selector_KQueue_Type = {
-	.wrap_struct_name = "IO_Event::Backend::KQueue",
+	.wrap_struct_name = "IO::Event::Backend::KQueue",
 	.function = {
 		.dmark = IO_Event_Selector_KQueue_Type_mark,
 		.dcompact = IO_Event_Selector_KQueue_Type_compact,
@@ -183,7 +164,7 @@ static const rb_data_type_t IO_Event_Selector_KQueue_Type = {
 		.dsize = IO_Event_Selector_KQueue_Type_size,
 	},
 	.data = NULL,
-	.flags = RUBY_TYPED_FREE_IMMEDIATELY,
+	.flags = RUBY_TYPED_FREE_IMMEDIATELY | RUBY_TYPED_WB_PROTECTED,
 };
 
 inline static
@@ -307,13 +288,17 @@ VALUE IO_Event_Selector_KQueue_allocate(VALUE self) {
 	struct IO_Event_Selector_KQueue *selector = NULL;
 	VALUE instance = TypedData_Make_Struct(self, struct IO_Event_Selector_KQueue, &IO_Event_Selector_KQueue_Type, selector);
 	
-	IO_Event_Selector_initialize(&selector->backend, Qnil);
+	IO_Event_Selector_initialize(&selector->backend, self, Qnil);
 	selector->descriptor = -1;
 	selector->blocked = 0;
 	
 	selector->descriptors.element_initialize = IO_Event_Selector_KQueue_Descriptor_initialize;
 	selector->descriptors.element_free = IO_Event_Selector_KQueue_Descriptor_free;
-	IO_Event_Array_allocate(&selector->descriptors, 1024, sizeof(struct IO_Event_Selector_KQueue_Descriptor));
+	
+	int result = IO_Event_Array_initialize(&selector->descriptors, IO_EVENT_ARRAY_DEFAULT_COUNT, sizeof(struct IO_Event_Selector_KQueue_Descriptor));
+	if (result < 0) {
+		rb_sys_fail("IO_Event_Selector_KQueue_allocate:IO_Event_Array_initialize");
+	}
 	
 	return instance;
 }
@@ -340,7 +325,7 @@ VALUE IO_Event_Selector_KQueue_initialize(VALUE self, VALUE loop) {
 	struct IO_Event_Selector_KQueue *selector = NULL;
 	TypedData_Get_Struct(self, struct IO_Event_Selector_KQueue, &IO_Event_Selector_KQueue_Type, selector);
 	
-	IO_Event_Selector_initialize(&selector->backend, loop);
+	IO_Event_Selector_initialize(&selector->backend, self, loop);
 	int result = kqueue();
 	
 	if (result == -1) {
@@ -371,7 +356,7 @@ VALUE IO_Event_Selector_KQueue_loop(VALUE self) {
 
 VALUE IO_Event_Selector_KQueue_idle_duration(VALUE self) {
 	struct IO_Event_Selector_KQueue *selector = NULL;
-	TypedData_Get_Struct(self, struct IO_Event_Selector_EPoll, &IO_Event_Selector_KQueue_Type, selector);
+	TypedData_Get_Struct(self, struct IO_Event_Selector_KQueue, &IO_Event_Selector_KQueue_Type, selector);
 	
 	double duration = selector->idle_duration.tv_sec + (selector->idle_duration.tv_nsec / 1000000000.0);
 	
@@ -396,7 +381,7 @@ VALUE IO_Event_Selector_KQueue_transfer(VALUE self)
 	struct IO_Event_Selector_KQueue *selector = NULL;
 	TypedData_Get_Struct(self, struct IO_Event_Selector_KQueue, &IO_Event_Selector_KQueue_Type, selector);
 	
-	return IO_Event_Selector_fiber_transfer(selector->backend.loop, 0, NULL);
+	return IO_Event_Selector_loop_yield(&selector->backend);
 }
 
 VALUE IO_Event_Selector_KQueue_resume(int argc, VALUE *argv, VALUE self)
@@ -420,7 +405,7 @@ VALUE IO_Event_Selector_KQueue_push(VALUE self, VALUE fiber)
 	struct IO_Event_Selector_KQueue *selector = NULL;
 	TypedData_Get_Struct(self, struct IO_Event_Selector_KQueue, &IO_Event_Selector_KQueue_Type, selector);
 	
-	IO_Event_Selector_queue_push(&selector->backend, fiber);
+	IO_Event_Selector_ready_push(&selector->backend, fiber);
 	
 	return Qnil;
 }
@@ -459,6 +444,7 @@ void process_prewait(pid_t pid) {
 		result = waitid(P_PID, pid, &info, WEXITED | WNOWAIT);
 		// This can sometimes get interrupted by SIGCHLD.
 	} while (result == -1 && errno == EINTR);
+	
 	if (result == -1) {
 		rb_sys_fail("process_prewait:waitid");
 	}
@@ -469,7 +455,7 @@ static
 VALUE process_wait_transfer(VALUE _arguments) {
 	struct process_wait_arguments *arguments = (struct process_wait_arguments *)_arguments;
 	
-	IO_Event_Selector_fiber_transfer(arguments->selector->backend.loop, 0, NULL);
+	IO_Event_Selector_loop_yield(&arguments->selector->backend);
 	
 	if (arguments->waiting->ready) {
 		process_prewait(arguments->pid);
@@ -503,6 +489,8 @@ VALUE IO_Event_Selector_KQueue_process_wait(VALUE self, VALUE fiber, VALUE _pid,
 		.events = IO_EVENT_EXIT,
 	};
 	
+	RB_OBJ_WRITTEN(self, Qundef, fiber);
+	
 	struct process_wait_arguments process_wait_arguments = {
 		.selector = selector,
 		.waiting = &waiting,
@@ -543,7 +531,7 @@ static
 VALUE io_wait_transfer(VALUE _arguments) {
 	struct io_wait_arguments *arguments = (struct io_wait_arguments *)_arguments;
 	
-	IO_Event_Selector_fiber_transfer(arguments->selector->backend.loop, 0, NULL);
+	IO_Event_Selector_loop_yield(&arguments->selector->backend);
 	
 	if (arguments->waiting->ready) {
 		return RB_INT2NUM(arguments->waiting->ready);
@@ -566,6 +554,8 @@ VALUE IO_Event_Selector_KQueue_io_wait(VALUE self, VALUE fiber, VALUE io, VALUE
 		.events = RB_NUM2INT(events),
 	};
 	
+	RB_OBJ_WRITTEN(self, Qundef, fiber);
+	
 	int result = IO_Event_Selector_KQueue_Waiting_register(selector, descriptor, &waiting);
 	if (result == -1) {
 		rb_sys_fail("IO_Event_Selector_KQueue_io_wait:IO_Event_Selector_KQueue_Waiting_register");
@@ -669,6 +659,8 @@ VALUE IO_Event_Selector_KQueue_io_read(VALUE self, VALUE fiber, VALUE io, VALUE
 		.offset = offset,
 	};
 	
+	RB_OBJ_WRITTEN(self, Qundef, fiber);
+	
 	return rb_ensure(io_read_loop, (VALUE)&io_read_arguments, io_read_ensure, (VALUE)&io_read_arguments);
 }
 
@@ -775,6 +767,8 @@ VALUE IO_Event_Selector_KQueue_io_write(VALUE self, VALUE fiber, VALUE io, VALUE
 		.offset = offset,
 	};
 	
+	RB_OBJ_WRITTEN(self, Qundef, fiber);
+	
 	return rb_ensure(io_write_loop, (VALUE)&io_write_arguments, io_write_ensure, (VALUE)&io_write_arguments);
 }
 
@@ -799,24 +793,21 @@ struct timespec * make_timeout(VALUE duration, struct timespec * storage) {
 		return NULL;
 	}
 	
-	if (FIXNUM_P(duration)) {
+	if (RB_INTEGER_TYPE_P(duration)) {
 		storage->tv_sec = NUM2TIMET(duration);
 		storage->tv_nsec = 0;
 		
 		return storage;
 	}
 	
-	else if (RB_FLOAT_TYPE_P(duration)) {
-		double value = RFLOAT_VALUE(duration);
-		time_t seconds = value;
-		
-		storage->tv_sec = seconds;
-		storage->tv_nsec = (value - seconds) * 1000000000L;
-		
-		return storage;
-	}
+	duration = rb_to_float(duration);
+	double value = RFLOAT_VALUE(duration);
+	time_t seconds = value;
 	
-	rb_raise(rb_eRuntimeError, "unable to convert timeout");
+	storage->tv_sec = seconds;
+	storage->tv_nsec = (value - seconds) * 1000000000L;
+	
+	return storage;
 }
 
 static
@@ -906,7 +897,7 @@ int IO_Event_Selector_KQueue_handle(struct IO_Event_Selector_KQueue *selector, u
 			IO_Event_List_append(node, saved);
 			
 			waiting->ready = matching_events;
-			IO_Event_Selector_fiber_transfer(waiting->fiber, 0, NULL);
+			IO_Event_Selector_loop_resume(&selector->backend, waiting->fiber, 0, NULL);
 			
 			node = saved->tail;
 			IO_Event_List_pop(saved);
@@ -963,7 +954,7 @@ VALUE IO_Event_Selector_KQueue_select(VALUE self, VALUE duration) {
 	selector->idle_duration.tv_sec = 0;
 	selector->idle_duration.tv_nsec = 0;
 	
-	int ready = IO_Event_Selector_queue_flush(&selector->backend);
+	int ready = IO_Event_Selector_ready_flush(&selector->backend);
 	
 	struct select_arguments arguments = {
 		.selector = selector,
@@ -984,7 +975,7 @@ VALUE IO_Event_Selector_KQueue_select(VALUE self, VALUE duration) {
 	// Non-comprehensive testing shows this gives a 1.5x speedup.
 	
 	// First do the syscall with no timeout to get any immediately available events:
-	if (DEBUG) fprintf(stderr, "\r\nselect_internal_with_gvl timeout=" PRINTF_TIMESPEC "\r\n", PRINTF_TIMESPEC_ARGS(arguments.storage));
+	if (DEBUG) fprintf(stderr, "\r\nselect_internal_with_gvl timeout=" IO_EVENT_TIME_PRINTF_TIMESPEC "\r\n", IO_EVENT_TIME_PRINTF_TIMESPEC_ARGUMENTS(arguments.storage));
 	select_internal_with_gvl(&arguments);
 	if (DEBUG) fprintf(stderr, "\r\nselect_internal_with_gvl done\r\n");
 	
@@ -1000,14 +991,14 @@ VALUE IO_Event_Selector_KQueue_select(VALUE self, VALUE duration) {
 			arguments.count = KQUEUE_MAX_EVENTS;
 			
 			struct timespec start_time;
-			IO_Event_Selector_current_time(&start_time);
+			IO_Event_Time_current(&start_time);
 			
-			if (DEBUG) fprintf(stderr, "IO_Event_Selector_KQueue_select timeout=" PRINTF_TIMESPEC "\n", PRINTF_TIMESPEC_ARGS(arguments.storage));
+			if (DEBUG) fprintf(stderr, "IO_Event_Selector_KQueue_select timeout=" IO_EVENT_TIME_PRINTF_TIMESPEC "\n", IO_EVENT_TIME_PRINTF_TIMESPEC_ARGUMENTS(arguments.storage));
 			select_internal_without_gvl(&arguments);
 			
 			struct timespec end_time;
-			IO_Event_Selector_current_time(&end_time);
-			IO_Event_Selector_elapsed_time(&start_time, &end_time, &selector->idle_duration);
+			IO_Event_Time_current(&end_time);
+			IO_Event_Time_elapsed(&start_time, &end_time, &selector->idle_duration);
 		}
 	}
 	
@@ -1054,9 +1045,27 @@ VALUE IO_Event_Selector_KQueue_wakeup(VALUE self) {
 	return Qfalse;
 }
 
+
+static int IO_Event_Selector_KQueue_supported_p(void) {
+	int fd = kqueue();
+	
+	if (fd < 0) {
+		rb_warn("kqueue() was available at compile time but failed at run time: %s\n", strerror(errno));
+		
+		return 0;
+	}
+	
+	close(fd);
+	
+	return 1;
+}
+
 void Init_IO_Event_Selector_KQueue(VALUE IO_Event_Selector) {
-	IO_Event_Selector_KQueue = rb_define_class_under(IO_Event_Selector, "KQueue", rb_cObject);
-	rb_gc_register_mark_object(IO_Event_Selector_KQueue);
+	if (!IO_Event_Selector_KQueue_supported_p()) {
+		return;
+	}
+	
+	VALUE IO_Event_Selector_KQueue = rb_define_class_under(IO_Event_Selector, "KQueue", rb_cObject);
 	
 	rb_define_alloc_func(IO_Event_Selector_KQueue, IO_Event_Selector_KQueue_allocate);
 	rb_define_method(IO_Event_Selector_KQueue, "initialize", IO_Event_Selector_KQueue_initialize, 1);
